package net.ihe.gazelle.proxy.netty;

import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import net.ihe.gazelle.proxy.netty.channel.ProxySide;
import net.ihe.gazelle.proxy.netty.channel.ServerPipelineFactory;

import org.jboss.netty.bootstrap.ServerBootstrap;
import org.jboss.netty.channel.Channel;
import org.jboss.netty.channel.ChannelHandler;
import org.jboss.netty.channel.ChannelPipelineFactory;
import org.jboss.netty.channel.socket.ClientSocketChannelFactory;
import org.jboss.netty.channel.socket.ServerSocketChannelFactory;
import org.jboss.netty.channel.socket.nio.NioClientSocketChannelFactory;
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory;

public abstract class Proxy<REQU, RESP> {

	private static final ExecutorService executor = Executors.newCachedThreadPool();

	// Set up the client socket factory (for connecting to remote host).
	public static final ClientSocketChannelFactory clientChannelFactory = new NioClientSocketChannelFactory(executor,
			executor);
	public static final ServerSocketChannelFactory serverChannelFactory = new NioServerSocketChannelFactory(executor,
			executor);

	private transient Channel serverChannel;
	private ConnectionConfig<REQU, RESP> connectionConfig;
	private transient ProxyEventListener<REQU, RESP> proxyEventListener;

	public Proxy(ProxyEventListener<REQU, RESP> proxyEventListener, ConnectionConfig<REQU, RESP> connectionConfig) {
		super();

		this.serverChannel = null;
		this.connectionConfig = connectionConfig;
		this.proxyEventListener = proxyEventListener;
	}

	public abstract REQU handleRequest(Object message);

	public abstract RESP handleResponse(Object message);

	public int getLocalPort() {
		return connectionConfig.getLocalPort();
	}

	public String getRemoteHost() {
		return connectionConfig.getRemoteHost();
	}

	public int getRemotePort() {
		return connectionConfig.getRemotePort();
	}

	public ChannelType getChannelType() {
		return connectionConfig.getChannelType();
	}

	public List<ChannelHandler> getRequestConnectionHandlers() {
		if (connectionConfig != null) {
			return connectionConfig.getRequestConnectionHandlers();
		} else {
			return new ArrayList<ChannelHandler>(0);
		}
	}

	public abstract List<ChannelHandler> getRequestDecoderChannels();

	public List<ChannelHandler> getResponseConnectionHandlers() {
		if (connectionConfig != null) {
			return connectionConfig.getResponseConnectionHandlers();
		} else {
			return new ArrayList<ChannelHandler>(0);
		}
	}

	public abstract List<ChannelHandler> getResponseDecoderChannels();

	public ChannelPipelineFactory getServerPipelineFactory() {
		return new ServerPipelineFactory<REQU, RESP>(this);
	}

	public final void handleMessage(Object message, ProxySide proxySide, String requesterIp, int requesterPort,
			Channel requestChannel) {
		switch (proxySide) {
		case REQUEST:
			REQU request = handleRequest(message);
			if (request != null) {
				proxyEventListener.onRequest(request, requesterIp, requesterPort, getLocalPort(), getRemoteHost(),
						getRemotePort(), requestChannel.getId());
			}
			break;
		case RESPONSE:
			RESP response = handleResponse(message);
			if (response != null) {
				proxyEventListener.onResponse(response, requesterIp, requesterPort, getLocalPort(), getRemoteHost(),
						getRemotePort(), requestChannel.getId());
			}
			break;
		}

	}

	public boolean isOpen() {
		if (serverChannel == null && serverChannel.isOpen()) {
			return true;
		}
		return false;
	}

	public void start() {
		if (serverChannel != null) {
			throw new IllegalStateException("Already started");
		}
		// Configure the TCP/IP server.
		ServerBootstrap bootstrap = new ServerBootstrap(serverChannelFactory);
		// Set up the event pipeline factory.
		bootstrap.setPipelineFactory(getServerPipelineFactory());
		// Bind and start to accept incoming connections.
		serverChannel = bootstrap.bind(new InetSocketAddress(getLocalPort()));
	}

	public void stop() {
		if (serverChannel == null) {
			throw new IllegalStateException("Not started");
		}
		serverChannel.close().awaitUninterruptibly();
		serverChannel = null;
	}

}
